/* CPU emulation loop and memory access
 *
 * (C) 2018 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <endian.h>
#include "execute.h"
#include "m68k.h"
#include "m68kcpu.h"

//#define DEBUG_CPU

#define IOMASK	0xfffff000
#define IOBASE	0x00dff000

static uint8_t *memory = NULL;
static uint32_t memory_size = 0;
static uint16_t *chipreg = NULL;

static uint16_t read_io(unsigned int address)
{
#ifdef DEBUG_CPU
	printf("Chip read from 0x%08x\n", address);
#endif
	return emulate_io_read(address);
}

static void write_io(unsigned int address, uint16_t value)
{
#ifdef DEBUG_CPU
	printf("Chip write to 0x%08x\n", address);
#endif
	if ((address & IOMASK) == IOBASE)
		chipreg[address & ~IOMASK] = value;
	emulate_io_write(address, value);
}

unsigned int m68k_read_memory_8(unsigned int address)
{
	if (address > memory_size - 1) {
		if ((address & 1)) {
			/* read lower (right) byte */
			return read_io(address & ~1);
		} else {
			/* read upper (left) byte */
			return read_io(address) >> 8;
		}
	}
	return *((uint8_t *)(memory + address));
}

unsigned int m68k_read_memory_16(unsigned int address)
{
	if (address > memory_size - 2) {
		return read_io(address);
	}
	return be16toh(*((uint16_t *)(memory + address)));
}

unsigned int m68k_read_memory_32(unsigned int address)
{
	if (address > memory_size - 4) {
		int32_t value;
		value = read_io(address) << 16;
		value |= read_io(address + 2);
		return value;
	}
	return be32toh(*((uint32_t *)(memory + address)));
}

void m68k_write_memory_8(unsigned int address, unsigned int value)
{
	if (address > memory_size - 1) {
		if ((address & 1)) {
			/* write lower (right) byte */
			write_io(address & ~1, 0xff00 | value);
		} else {
			/* write upper (left) byte */
			write_io(address, 0x00ff | (value << 8));
		}
		return;
	}
	*((uint8_t *)(memory + address)) = value;
}

void m68k_write_memory_16(unsigned int address, unsigned int value)
{
	if (address > memory_size - 2) {
		write_io(address, value);
		return;
	}
	*((uint16_t *)(memory + address)) = htobe16(value);
}

void m68k_write_memory_32(unsigned int address, unsigned int value)
{
	if (address > memory_size - 4) {
		write_io(address, value >> 16);
		write_io(address + 2, value);
		return;
	}
	*((uint32_t *)(memory + address)) = htobe32(value);
}

void execute_init(int32_t _memory_size, uint8_t *_memory, uint16_t *_chipreg)
{
	memory_size = _memory_size;
	memory = _memory;
	chipreg = _chipreg;

	/* init CPU */
	m68k_init();
        m68k_set_cpu_type(M68K_CPU_TYPE_68000);
}

void reset_cpu(void)
{
	m68k_pulse_reset();
}

int execute_cpu(int irq, const struct cpu_stop stop_at[], int *event)
{
	int cycle_count = 0;
	int instruction_count = 0;
	int i;
	uint32_t save_pc;

	if (irq) {
		save_pc = REG_PC;
		m68k_set_irq(irq);
		/* loop until we returned from interrupt */
		do {
			/* execute one opcode */
			cycle_count += m68k_execute(0);
			instruction_count++;
		} while (REG_PC != save_pc);
#ifdef DEBUG_CPU
		printf("interrupt %d took %d opcodes\n", irq, instruction_count);
#endif
	} else {
		do {
			/* execute one opcode */
			cycle_count += m68k_execute(0);
			instruction_count++;

			/* checking if game does not hit any 'stop_at' break point, give error output */
			if (instruction_count >= 10000000) {
				if (instruction_count == 10000000)
					fprintf(stderr, "!!! games seems to got stuck in an endless loop, please fix !!!\n");
				fprintf(stderr, "program counter at: %06x\n", REG_PC);
				if (instruction_count == 10000020)
					break;
			}

			for (i = 0; stop_at[i].event; i++) {
				if (REG_PC == stop_at[i].pc) {
#ifdef DEBUG_CPU
					printf("execution to address 0x%06x took %d opcodes\n", REG_PC, instruction_count);
#endif
					*event = stop_at[i].event;
					return cycle_count;
				}
			}
		} while (42);
	}

	return cycle_count;
}

