/* disk access emulation
 *
 * (C) 2018 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include "disk.h"

//#define DEBUG_DISK

#define CIAAPRA		0xbfe000
#define CIABPRB		0xbfd100
#define DSKPTH		0xdff020
#define DSKPTL		0xdff022
#define DSKLEN		0xdff024
#define INTREQR		0xdff01e

static uint16_t ciabprb = 0xffff;
static int track = 0;
static uint32_t dskptr = 0;
static uint16_t dsklen = 0;

/* NOTE: all CIA access is WORD access, so we use bits 8..15 for CIA-B */

uint16_t emulate_disk_read(uint32_t address)
{
	uint16_t value = 0xffff;

	switch (address) {
	case CIAAPRA:
		/* /RDY (bit 5) is always 0, so disk is ready! */
		value &= 0xffdf;
		if (track == 0)
			value &= 0xffef;
		break;
	case CIABPRB:
		/* return current bits from CIA-A PRB */
		value &= ciabprb;
		break;
	case INTREQR:
		/* if disk dma is not on, don't process read/write */
		if (!(dsklen & 0x8000))
			break;

		if ((dsklen & 0x4000))
			disk_write(track, (ciabprb >> 10) & 1, dskptr, (dsklen & 0x3fff) << 1);
		else
			disk_read(track, (ciabprb >> 10) & 1, dskptr, (dsklen & 0x3fff) << 1);
#ifdef DEBUG_DISK
		printf("read from INTREQR: game asks for finished read/write\n");
#endif
		/* keep bit 2 set, so it indicated finished read/write */
		value &= 0xffff;
		break;
	}

	return value;
}

void emulate_disk_write(uint32_t address, uint16_t value)
{
	switch (address) {
	case CIABPRB:
#ifdef DEBUG_DISK
		printf("selected side %d\n", (value >> 10) & 1);
#endif
		if (((ciabprb >> 8) & 1) && !((value >> 8) & 1)) {
			if (!((value >> 9) & 1)) {
				track++;
				if (track > 84)
					track = 84;
#ifdef DEBUG_DISK
				printf("step forward to track %d\n", track);
#endif
			} else {
				track--;
				if (track < 0)
					track = 0;
#ifdef DEBUG_DISK
				printf("step backwards to track %d\n", track);
#endif
			}
		}
		ciabprb = value;
		break;
	case DSKPTH:
		dskptr = (dskptr & 0x0000ffff) | (value << 16);
		break;
	case DSKPTL:
		dskptr = (dskptr & 0xffff0000) | value;
#ifdef DEBUG_DISK
		printf("disk pointer set to 0x%06x\n", dskptr);
#endif
		break;
	case DSKLEN:
		dsklen = value;
#ifdef DEBUG_DISK
		printf("DSKLEN: disk length set to 0x%04x, mode set to '%s', dma set to '%s'\n", dsklen & 0x3fff, (dsklen & 0x4000) ? "write" : "read", (dsklen & 0x8000) ? "on" : "off");
#endif
		break;
	}
}

int get_disk_track(void)
{
#ifdef DEBUG_DISK
	printf("disk drive is at track %d\n", track);
#endif
	return track;
}

