/* keyboard emulation
 *
 * (C) 2018 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include "keyboard.h"

//#define DEBUG_KEYS

/* stolen from UAE */
static struct keycodes {
	const char *key;
	uint8_t code;
} keycodes[] = {
	{ "A", 0x20 }, { "B", 0x35 }, { "C", 0x33 }, { "D", 0x22 },
	{ "E", 0x12 }, { "F", 0x23 }, { "G", 0x24 }, { "H", 0x25 },
	{ "I", 0x17 }, { "J", 0x26 }, { "K", 0x27 }, { "L", 0x28 },
	{ "M", 0x37 }, { "N", 0x36 }, { "O", 0x18 }, { "P", 0x19 },
	{ "Q", 0x10 }, { "R", 0x13 }, { "S", 0x21 }, { "T", 0x14 },
	{ "U", 0x16 }, { "V", 0x34 }, { "W", 0x11 }, { "X", 0x32 },
	{ "Y", 0x15 }, { "Z", 0x31 },

	{ "0", 0x0A }, { "1", 0x01 }, { "2", 0x02 }, { "3", 0x03 },
	{ "4", 0x04 }, { "5", 0x05 }, { "6", 0x06 }, { "7", 0x07 },
	{ "8", 0x08 }, { "9", 0x09 },

	{ "NP0", 0x0F }, { "NP1", 0x1D }, { "NP2", 0x1E }, { "NP3", 0x1F },
	{ "NP4", 0x2D }, { "NP5", 0x2E }, { "NP6", 0x2F }, { "NP7", 0x3D },
	{ "NP8", 0x3E }, { "NP9", 0x3F },

	{ "NPDIV", 0x5C }, { "NPMUL", 0x5D }, { "NPSUB", 0x4A },
	{ "NPADD", 0x5E }, { "NPDEL", 0x3C }, { "NPLPAREN", 0x5A },
	{ "NPRPAREN", 0x5B },

	{ "F1", 0x50 }, { "F2", 0x51 }, { "F3", 0x52 }, { "F4", 0x53 },
	{ "F5", 0x54 }, { "F6", 0x55 }, { "F7", 0x56 }, { "F8", 0x57 },
	{ "F9", 0x58 }, { "F10", 0x59 },

	{ "UP", 0x4C }, { "DN", 0x4D }, { "LF", 0x4F }, { "RT", 0x4E },

	{ "SPC", 0x40 }, { "BS", 0x41 }, { "TAB", 0x42 }, { "ENT", 0x43 },
	{ "RET", 0x44 }, { "ESC", 0x45 }, { "DEL", 0x46 },

	{ "LSH", 0x60 }, { "RSH", 0x61 }, { "CAPSLOCK", 0x62 },
	{ "CTRL", 0x63 }, { "LALT", 0x64 }, { "RALT", 0x65 }, { "LAMI", 0x66 },
	{ "RAMI", 0x67 }, { "HELP", 0x5F },

	{ "LBRACKET", 0x1A }, { "RBRACKET", 0x1B }, { "SEMICOLON", 0x29 },
	{ "COMMA", 0x38 }, { "PERIOD", 0x39 }, { "SLASH", 0x3A },
	{ "BACKSLASH", 0x0D }, { "QUOTE", 0x2A }, { "NUMBERSIGN", 0x2B },
	{ "LTGT", 0x30 }, { "BACKQUOTE", 0x00 }, { "MINUS", 0x0B },
	{ "EQUAL", 0x0C },

	{ NULL, 0 },
};

static uint8_t buffer[16];
static int buffer_size = sizeof(buffer), buffer_len = 0;
static uint32_t keydown[4] = {0, 0, 0, 0};

#define ROL(v)		((v << 1) | (v >> 7))

#define CIAASDR		0xbfec00
#define CIAAICR		0xbfed00
#define CIAACRA		0xbfee00

uint16_t emulate_keyboard_read(uint32_t address)
{
	uint8_t value;
	int i;

	switch (address) {
	case CIAAICR:
		if (buffer_len)
			return 0xff08;
		return 0xff00;
	case CIAASDR:
		if (buffer_len == 0)
			return 0xffff;
		value = buffer[--buffer_len];
		/* rotate and negate */
		value = ~ROL(value);
#ifdef DEBUG_KEYS
		printf("sending key code %02x\n", buffer[buffer_len]);
#endif
		for (i = 0; i < buffer_len; i++)
			buffer[i] = buffer[i + 1];
		return 0xff00 | value;
	}

	return 0xffff;
}

void set_key(const char *key, int down)
{
	int code = -1, isdown;
	int i;

	for (i = 0; keycodes[i].key; i++) {
		if (!strcmp(keycodes[i].key, key)) {
			code = keycodes[i].code;
			break;
		}
	}
	if (code < 0 || code >= 0x80) {
		fprintf(stderr, "Key code '%s' unknown, please fix!\n", key);
		return;
	}

	isdown = (keydown[code >> 5] >> (code & 0x1f)) & 1;
	if (down) {
		if (isdown) {
//			printf("key already down\n");
			return;
		}
		keydown[code >> 5] |= (1 << (code & 0x1f));
	} else {
		if (!isdown) {
//			printf("key already up\n");
			return;
		}
		keydown[code >> 5] &= ~(1 << (code & 0x1f));
		code |= 0x80;
	}

	if (buffer_len == buffer_size) {
		fprintf(stderr, "keyboard buffer overflow\n");
		return;
	}

	buffer[buffer_len++] = code;
#ifdef DEBUG_KEYS
	printf("Converting key '%s' to code 0x%02x\n", key, code);
#endif
}


