/* OpenGL rendering
 *
 * (C) 2018 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <errno.h>
#include "opengl.h"
#include <GL/glew.h>

static uint8_t *text_rgb = NULL;
static GLuint text_name;
static int screen_width, screen_height;
static int image_width, image_height;
static int texture_size;

/* alloc and init an image texture with size that is greater than the power of two */
int init_opengl(int _image_width, int _image_height)
{
	int rc;

	image_width = _image_width;
	image_height = _image_height;

	/* generate texture */
	for (texture_size = 1; texture_size <= image_width || texture_size <= image_height; texture_size *= 2)
		;
	text_rgb = calloc(texture_size * texture_size * 10, 3);
	if (!text_rgb) {
		fprintf(stderr, "Failed to allocate texture\n");
		rc = -ENOMEM;
		goto error;
	}
	glClearColor(0.0, 0.0, 0.0, 1.0);
	glShadeModel(GL_FLAT);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); /* bytes */
	glGenTextures(1, &text_name);
	glBindTexture(GL_TEXTURE_2D, text_name);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, texture_size, texture_size, 0, GL_RGB, GL_UNSIGNED_BYTE, text_rgb);

	return 0;

error:
	return rc;
}

/* set clip planes so that the image portion of the image texture is centered and pixels are rectengular */
void resize_opengl(int _screen_width, int _screen_height)
{
	double width_border = 1.0;
	double height_border = 1.0;

	if (_screen_width < 1 || _screen_height < 1)
		return;
	screen_width = _screen_width;
	screen_height = _screen_height;

	if (image_width * screen_height > screen_width * image_height) {
		height_border = (double)(image_width * screen_height) / (double)(screen_width * image_height);
	}
	if (image_width * screen_height < screen_width * image_height) {
		width_border = (double)(screen_width * image_height / (double)(image_width * screen_height));
	}

	/* viewport and projection matrix */
	glViewport(0, 0, (GLsizei)screen_width, (GLsizei)screen_height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	double width = (double)image_width / (double)texture_size;
	double height = (double)image_height / (double)texture_size;
	glOrtho(
			-width * (width_border - 1.0) / 2,
			width / 2 + width * width_border / 2,
			height / 2 + height * height_border / 2,
			-height * (height_border - 1.0) / 2,
		       	-1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
}

/* render image texture */
void render_opengl(uint8_t *rgb, int filter)
{
	glClear(GL_COLOR_BUFFER_BIT);
	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);  /* no modulation with color */
	glBindTexture(GL_TEXTURE_2D, text_name);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, (filter) ? GL_LINEAR : GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, (filter) ? GL_LINEAR : GL_NEAREST);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, image_width, image_height, GL_RGB, GL_UNSIGNED_BYTE, rgb);
	glBegin(GL_QUADS);
	glTexCoord2f(0.0, 0.0);
	glVertex3f(0.0, 0.0, 0.0);
	glTexCoord2f(1.0, 0.0);
	glVertex3f(1.0, 0.0, 0.0);
	glTexCoord2f(1.0, 1.0);
	glVertex3f(1.0, 1.0, 0.0);
	glTexCoord2f(0.0, 1.0);
	glVertex3f(0.0, 1.0, 0.0);
	glEnd();
	glDisable(GL_TEXTURE_2D);
}

/* free image texture */
void exit_opengl(void)
{
	if (text_rgb) {
		free(text_rgb);
		text_rgb = NULL;
	}
}

