/* sound emulation
 *
 * (C) 2018 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include "sound.h"
#include "../libcpu/m68kcpu.h"

//#define DEBUG_SOUND

#define CLOCK		3546895.0
#define DMACON		0xdff096
#define IOBASE		0xdff000
#define AUDIOBASE	0xdff0a0
#define AUDxLCH		0x0
#define AUDxLCL		0x2
#define AUDxLEN		0x4
#define AUDxPER		0x6
#define AUDxVOL		0x8

static struct audio_channel {
	int		dma_on;		/* set if dma is running */
	uint32_t	start;		/* sample address to start from */
	int		length;		/* length of samples (not words) */
	uint32_t	pointer;	/* pointer of renderer */
	uint32_t	end;		/* end of current sample */
	double		offset;		/* offset between two (input) samples */
	double		step;		/* number of input samples per output sample */
	double		volume;		/* volume of channel in the result sample */
	float		sample;		/* current (last read) sample */
} channel[4] = {
	{ 0, 0xffffffff, 0, 0, 0, 0.0, 0.0, 0.0, 0.0 },
	{ 0, 0xffffffff, 0, 0, 0, 0.0, 0.0, 0.0, 0.0 },
	{ 0, 0xffffffff, 0, 0, 0, 0.0, 0.0, 0.0, 0.0 },
	{ 0, 0xffffffff, 0, 0, 0, 0.0, 0.0, 0.0, 0.0 },
};

double filter_alpha = 0.0, filter_last = 0.0;

static int start_dma(int c, struct audio_channel *chan)
{
	/* fetch sample start and length and use it for rendering */
	chan->pointer = chan->start;
	chan->end = chan->start + chan->length;
	/* check ranges */
	if (chan->length == 0) {
		fprintf(stderr, "Sample length of channel %d is 0, please fix!\n", c);
		chan->dma_on = 0;
		return -1;
	}
	if (chan->pointer >= 0x80000) {
		fprintf(stderr, "Sample pointer 0x%08x of channel %d is outside memory, please fix!\n", chan->pointer, c);
		chan->dma_on = 0;
		return -1;
	}
	if (chan->end > 0x80000) {
		fprintf(stderr, "Sample end (pointer 0x%08x + length 0x%x) of channel %d is outside memory, please fix!\n", chan->pointer, chan->length, c);
		chan->dma_on = 0;
		return -1;
	}

	return 0;
}

void emulate_sound_write(uint32_t address, uint16_t value, int samplerate)
{
	int c;

	if (address == DMACON) {
		/* handle DMA flag and start DMA, if required */
		for (c = 0; c < 4; c++) {
			/* if nothing should be changed */
			if (!(value & (1 << c)))
			 	continue;
			if (!(value & 0x8000)) {
				channel[c].dma_on = 0;
#ifdef DEBUG_SOUND
				printf("Audio channel %d: Turning DMA off\n", c);
#endif
				continue;
			}
			if (channel[c].dma_on == 1)
				continue;
			if (start_dma(c, &channel[c]) < 0) {
#ifdef DEBUG_SOUND
				printf("Audio channel %d: Turning DMA on failed\n", c);
#endif
				continue;
			}
#ifdef DEBUG_SOUND
			printf("Audio channel %d: Turning DMA on\n", c);
#endif
			channel[c].dma_on = 1;
		}
		return;
	}

	/* return if not register OR get channel from address */
	if (address < AUDIOBASE || address >= AUDIOBASE + 0x03f)
		return;
	c = (address - AUDIOBASE) >> 4;

	switch (address & 0xf) {
	case AUDxLCH:
#ifdef DEBUG_SOUND
		printf("Audio channel %d: Setting high word of sample start to 0x%04x\n", c, value);
#endif
		channel[c].start = (channel[c].start & 0x0000ffff) | (value << 16);
		break;
	case AUDxLCL:
#ifdef DEBUG_SOUND
		printf("Audio channel %d: Setting low word of sample start to 0x%04x\n", c, value);
#endif
		channel[c].start = (channel[c].start & 0xffff0000) | (value & 0xfffe);
		break;
	case AUDxLEN:
#ifdef DEBUG_SOUND
		printf("Audio channel %d: Setting sample length to %d words\n", c, value);
#endif
		channel[c].length = value << 1;
		break;
	case AUDxPER:
		if (value == 0)
			value = 1;
#ifdef DEBUG_SOUND
		printf("Audio channel %d: Setting sample frequency to %.2f Hz\n", c, CLOCK / (double)value);
#endif
		channel[c].step = CLOCK / (double)value / (double)samplerate;
		/* sample rate never at or above output rate */
		if (channel[c].step > 0.999)
			channel[c].step = 0.999;
		break;
	case AUDxVOL:
		if (value > 64)
			value = 64;
#ifdef DEBUG_SOUND
		printf("Audio channel %d: Setting volume to %d (0..64)\n", c, value);
#endif
		channel[c].volume = (double)value / 64.0 / 4.0; /* 1/4th volume per channel */
		break;
	}
}

/* render sound from memory and add each channel to ring buffer array */
void render_sound(uint8_t *memory, float *sample, int buffer_size, int buffer_pos, int length, int filter)
{
	int c, s, pos;
	struct audio_channel *chan;
	double new_offset;
	float new_sample;

	/* clear buffer first */
	pos = buffer_pos;
	for (s = 0; s < length; s++) {
		sample[pos] = 0.0;
		pos = (pos + 1) % buffer_size;
	}

	/* render each channel and sum the result */
	for (c = 0; c < 4; c++) {
		chan = &channel[c];
		/* skip if DMA is disabled */
		if (!chan->dma_on)
			continue;
		/* Note: We must render sound, even if volume is 0.0,
		 * because the sample pointers must continue, as it would happen in a real Amiga.
		 * But this may have no practical use in this game.
		 */
		pos = buffer_pos;
		for (s = 0; s < length; s++) {
			new_offset = chan->offset + chan->step;
			if (new_offset < 1.0)
				/* no interpolation: use last input sample, because there is no change */
				sample[pos] += chan->sample * chan->volume;
			else {
				/* there is a change of input sample, so read new sample */
				new_offset -= 1.0;
				new_sample = (double)(((int8_t *)memory)[chan->pointer++]) / 128.0;
				if (chan->pointer == chan->end) {
					if (start_dma(c, chan) < 0)
						break;
				}
				/* linear interpolation: add portion of last sample to portion of new sample */
				sample[pos] += (chan->sample * (1.0 - chan->offset) + new_sample * new_offset) / chan->step * chan->volume;
				chan->sample = new_sample;
			}
			chan->offset = new_offset;
			pos = (pos + 1) % buffer_size;
		}
	}

	if (filter) {
		/* do low-pass filtering */
		pos = buffer_pos;
		for (s = 0; s < length; s++) {
			/* y[i] := y[i-1] + alpha * (x[i] - y[i-1]) */
			filter_last = sample[pos] = filter_last + filter_alpha * (sample[pos] - filter_last);
			pos = (pos + 1) % buffer_size;
		}
	}
}

void sound_init_filter(int samplerate)
{
	double dt = 1.0 / (double)samplerate;
	double C = 0.0000001;
	double R = 360;
	double alpha = dt / (R*C + dt);
#ifdef DEBUG_SOUND
	double cutoff = 1.0 / (2 * M_PI * R * C);
#endif

#ifdef DEBUG_SOUND
	printf("filter-cutoff: %.0f Hz (alpha = %.4f)\n", cutoff, alpha);
#endif
	filter_alpha = alpha;
}

