/* video emulation
 *
 * (C) 2018 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include "video.h"
#include "../libcpu/m68kcpu.h"

//#define DEBUG_COPPERLIST

static void planar2chunky(uint8_t *rgb, uint8_t *bitplanes[], uint16_t palette[], int planes, int width, int y_start, int y_end)
{
	int x, y, p, b, i;
	uint8_t word[planes], chunk;
	uint16_t rgb4;

	/* we start memory read from the following calulated bitplane offset: */
	i = y_start * (width / 8);
	rgb += y_start * width * 3;
	for (y = y_start; y < y_end; y++) {
		for (x = 0; x < width; x += 8) {
			for (p = 0; p < planes; p++)
				word[p] = bitplanes[p][i];
			for (b = 0; b < 8; b++) {
				chunk = (word[planes - 1] >> 7);
				word[planes - 1] <<= 1;
				for (p = planes - 2; p >= 0; p--) {
					chunk = (chunk << 1) | (word[p] >> 7);
					word[p] <<= 1;
				}
				rgb4 = palette[chunk];
				*rgb++ = (rgb4 >> 4) & 0xf0;
				*rgb++ = rgb4 & 0xf0;
				*rgb++ = rgb4 << 4;
			}
			i++;
		}
	}
}

static void planar2chunky2(uint8_t *rgb1, uint8_t *bitplanes[], uint16_t palette[], int planes, int width, int y_start, int y_end)
{
	uint8_t *rgb2, red, green, blue;
	int x, y, p, b, i;
	uint8_t word[planes], chunk;
	uint16_t rgb4;

	rgb2 = rgb1 + width * 6;

	/* we start memory read from the following calulated bitplane offset: */
	i = y_start * (width / 8);
	rgb1 += 2 * y_start * width * 6;
	rgb2 += 2 * y_start * width * 6;
	for (y = y_start; y < y_end; y++) {
		for (x = 0; x < width; x += 8) {
			for (p = 0; p < planes; p++)
				word[p] = bitplanes[p][i];
			for (b = 0; b < 8; b++) {
				chunk = (word[planes - 1] >> 7);
				word[planes - 1] <<= 1;
				for (p = planes - 2; p >= 0; p--) {
					chunk = (chunk << 1) | (word[p] >> 7);
					word[p] <<= 1;
				}
				rgb4 = palette[chunk];
				red = (rgb4 >> 4) & 0xf0;
				green = rgb4 & 0xf0;
				blue = rgb4 << 4;
				*rgb1++ = red;
				*rgb1++ = green;
				*rgb1++ = blue;
				*rgb1++ = red;
				*rgb1++ = green;
				*rgb1++ = blue;
				*rgb2++ = red;
				*rgb2++ = green;
				*rgb2++ = blue;
				*rgb2++ = red;
				*rgb2++ = green;
				*rgb2++ = blue;
			}
			i++;
		}
		rgb1 += width * 6;
		rgb2 += width * 6;
	}
}

#define COP1LCH		0x080
#define COP1LCL		0x082
#define COP2LCH		0x084
#define COP2LCL		0x086
#define BPL1PTH		0x0e0
#define COLOR00		0x180

void emul_video(uint8_t *rgb, uint8_t *memory, uint16_t render_palette[], int width, int height, int diwstart, uint16_t *io, int start, int stop, int double_size)
{
	uint32_t bitplane[8] = {0, 0, 0, 0, 0, 0, 0, 0};
	uint8_t *bitmem[8];
	uint16_t palette[16];
	uint32_t copperlist;
	uint16_t c1, c2;
	int row, last_row, line, last_line, from, to;
	int count;
	int i;
	int all_white;

	/* special case where all palette entries are white. (unknown reason, maybe due to teleporter travel) */
	all_white = 1;
	for (i = 0; i < 16; i++) {
		if (io[i * 2 + 0x180] != 0xfff)
			all_white = 0;
	}

	/* First set palette as specified in IO space.
	 * This is set with last IRQ, so it is used after VBL.
	 * Later the copper list causes them to be changed */
	for (i = 0; i < 16; i++) {
		/* use palette if all value are white, else use the palette that is used during VBL-IRQ */
		if (all_white)
			palette[i] = io[i * 2 + 0x180];
		else {
			/* use palette that will be set during IRQ routine.
			 * we need to copy it here, because we don't call IRQ routine for every rendered frame,
			 * so the color registers are not set by IRQ routine for every frame.
			 */
			palette[i] = render_palette[i];
		}
	}

	/* get copper list start pointer */
	copperlist = (io[COP1LCH] << 16) | io[COP1LCL];
	if (!copperlist) {
		fprintf(stderr, "Copper list pointer not initialized, please fix!\n");
		return;
	}

#ifdef DEBUG_COPPERLIST
	printf("Copper list reading from pointer: %06x\n", copperlist);
#endif

	/* parse copper list */
	count = 0;
	last_row = 0;
	last_line = 0;
	while (42) {
		if (++count == 100) {
			fprintf(stderr, "Copper list does not seem to terminate, please fix!\n");
			return;
		}
		c1 = m68k_read_memory_16(copperlist);
		c2 = m68k_read_memory_16(copperlist + 2);
#ifdef DEBUG_COPPERLIST
		printf("%06x: C1=%04x C2=%04x\n", copperlist, c1, c2);
#endif
		copperlist += 4;
		if (c1 == 0x0088) {
			copperlist = (io[COP1LCH] << 16) | io[COP1LCL];
#ifdef DEBUG_COPPERLIST
			printf("switching to 1st copperlist=%06x\n", copperlist);
#endif
			continue;
		}
		if (c1 == 0x008a) {
			copperlist = (io[COP2LCH] << 16) | io[COP2LCL];
#ifdef DEBUG_COPPERLIST
			printf("switching to 2nd copperlist=%06x\n", copperlist);
#endif
			continue;
		}
		if (!(c1 & 1)) {
			/* MOVE */
#ifdef DEBUG_COPPERLIST
			printf("MOVE 0xdff%03x = 0x%04x\n", c1, c2);
#endif
			/* get bitplane pointers */
			if (c1 >= BPL1PTH && c1 <= BPL1PTH + 32) {
				if ((c1 & 2))
					bitplane[(c1 - BPL1PTH) / 4] = (bitplane[(c1 - BPL1PTH) / 4] & 0xffff0000) | c2;
				else
					bitplane[(c1 - BPL1PTH) / 4] = (bitplane[(c1 - BPL1PTH) / 4] & 0x0000ffff) | (c2 << 16);
			}
			/* get color registers */
			if (c1 >= COLOR00 && c1 <= COLOR00 + 32) {
				palette[(c1 - COLOR00) / 2] = c2;
			}
		} else {
			if ((c2 & 1)) {
				fprintf(stderr, "We suppport no SKIP command in copper list, please fix!\n");
				continue;
			}
			/* WAIT */
			/* get new raster position.
			 * if the value is lower or equal, we add 256 lines
			 * we ignore column, since it is not relevant for this game.
			 */
			row = (c1 >> 8) & ((c2 >> 8) | 0x80);
			if (row < (last_row & 0xff))
				row |= 0x100;
#ifdef DEBUG_COPPERLIST
			printf("WAIT row = 0x%02x & 0x%02x = %d\n", c1 >> 8, (c2 >> 8) | 0x80, row);
#endif
			last_row = row;
			/* line relative to display window start */
			line = row - diwstart;
			/* continue: before display window start */
			if (line <= last_line)
				continue;
			/* render up to height */
			if (line >= height)
				line = height;
			/* check if bitplane pointers are set */
#ifdef DEBUG_COPPERLIST
			printf("Bitplanes:");
#endif
			for (i = 0; i < 4; i++) {
#ifdef DEBUG_COPPERLIST
				printf(" %06x", bitplane[i]);
#endif
				if (bitplane[i] == 0 || bitplane[i] + width * height / 8 >= 0x80000) {
					printf("\n");
					fprintf(stderr, "Bitplane %d in copper list not set or out of range, please fix!\n", i);
					return;
				}
				bitmem[i] = memory + bitplane[i];
			}
#ifdef DEBUG_COPPERLIST
			printf("\n");
#endif
			/* render portion before WAIT line */
#ifdef DEBUG_COPPERLIST
			printf("Render from line %d to line %d\n", last_line, line - 1);
#endif
#ifdef DEBUG_COPPERLIST
			printf("Palette:");
			for (i = 0; i < 16; i++) {
				printf(" %03x", palette[i] & 0xfff);
			}
			printf("\n");
#endif

			from = last_line;
			to = line;
			if (start > from)
				from = start;
			if (stop < to)
				to = stop;
			if (to > from) {
				if (double_size)
					planar2chunky2(rgb, bitmem, palette, 4, width, from, to);
				else
					planar2chunky(rgb, bitmem, palette, 4, width, from, to);
			}
			/* done if we rendered up to height */
			if (line == height)
				break;
			/* remeber for next wait command */
			last_row = row;
			last_line = line;
		}
	}

}
