/* main routine
 *
 * (C) 2018 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <endian.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "../libsdl/sdl.h"
#include "../libsdl/opengl.h"
#include "../libcpu/m68k.h"
#include "../libcpu/execute.h"
#include "../libvideo/video.h"
#include "../libsound/sound.h"
#include "../libjoystick/joystick.h"
#include "../libkeyboard/keyboard.h"
#include "../libdisk/disk.h"
#include "mercenary.h"

static int config_amiga_speed = 1;
static const char *config_gamesave_dir = ".mercenary";
static int config_video_filter = 1;
static int config_audio_filter = 1;

#define CPU_SPEED	7093790.0;

#define SOUND_SAMPLERATE 48000
/* - game IRQ rate
 * - must be used for sound rendering chunk
 */
#define IRQ_RATE	50
/* numbe of buffer chunks to dejitter:
 * - SDL render interval
 * - sound rendering interval
 * - sound card interval
 * 4 should be minimum, if video rate is >=50 Hz
 */
#define SOUND_CHUNKS	4

/* test sound to check if buffer does not overflow / underrun */
//#define DEBUG_SOUND_BUFFERING

#define IOSIZE		0x1000 /* bytes in io space */
#define	MEMORY_SIZE	0x80000
static uint8_t *memory = NULL;
static uint8_t *image = NULL;
#define SCREEN_WIDTH	(320*3)
#define SCREEN_HEIGHT	(200*3)
#define IMAGE_WIDTH	320
#define IMAGE_HEIGHT	200
#define BENSON_AT_LINE	136
#define IMAGE_DIWSTART	0x2c
static uint16_t *chipreg = NULL;
static float *sound_buffer = NULL; /* sound buffer memory */
static int sound_buffer_size; /* buffer sample size */
static int sound_buffer_writep; /* write pointer at buffer */
static int sound_buffer_readp; /* read pointer at buffer */
static int double_size = 1; /* render in double size, so each pixle is 2*2 pixles wide */

static const char *home_dir;

static int quit = 0;

int parse_args(int argc, char *argv[])
{
	int i = 1;

	while (argc > i) {
		if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
			printf("Usage: %s\n", argv[0]);
			printf(" -s --amiga-speed original | full\n");
			printf("        Set speed of rendering to original Amiga or full speed.\n");
			printf(" -v --video-filter on | off\n");
			printf("        Set video filter.\n");
			printf(" -a --audio-filter on | off\n");
			printf("        Set audio filter.\n");
			return -1;
		} else
		if (!strcmp(argv[i], "-s") || !strcmp(argv[i], "--amiga-speed")) {
			i++;
			if (argc == i) {
missing_parameter:
				printf("Missing parameter, use '--help'!\n");
				return -1;
			}
			if (!strcmp(argv[i], "original"))
				config_amiga_speed = 1;
			else
			if (!strcmp(argv[i], "full"))
				config_amiga_speed = 0;
			else {
illegal_parameter:
				printf("Illegal parameter, use '--help'!\n");
				return -1;
			}
		} else
		if (!strcmp(argv[i], "-v") || !strcmp(argv[i], "--video-filter")) {
			i++;
			if (argc == i)
				goto missing_parameter;
			if (!strcmp(argv[i], "on"))
				config_video_filter = 1;
			else
			if (!strcmp(argv[i], "off"))
				config_video_filter = 0;
			else
				goto illegal_parameter;
		} else
		if (!strcmp(argv[i], "-a") || !strcmp(argv[i], "--audio-filter")) {
			i++;
			if (argc == i)
				goto missing_parameter;
			if (!strcmp(argv[i], "on"))
				config_audio_filter = 1;
			else
			if (!strcmp(argv[i], "off"))
				config_audio_filter = 0;
			else
				goto illegal_parameter;
		} else {
			printf("Illegal option '%s', use '--help'!\n", argv[i]);
			return -1;
		}
		i++;
	}

	return 0;
}

static void special_event(int event)
{
}

static void main_loop(void)
{
	int had_first_irq = 0;
	static uint32_t current_time, last_time = 0, diff;
	int i, rc;
	int space, length;
	int cycle_count, event;
	double render_delay = 0.0;
	uint32_t palette_address;
	uint16_t palette[16];

	last_time = SDL_GetTicks();

	/* render result on window */
	while (!quit) {
		/* STEP 1: let the CPU render/process the game */
		/* don't render if we still delay */
		if (!render_delay) {
			/* execute until the rendered image is ready (wait for VBL) */
			cycle_count = 0;
			do {
				cycle_count += execute_cpu(0, mercenary_stop_at, &event);
				/* handle special events */
				if (event != STOP_AT_WAIT_VBL)
					special_event(event);
			} while (event != STOP_AT_WAIT_VBL);
			/* copy palette */
			palette_address = mercenary_palette();
			for (i = 0; i < 16; i++)
				palette[i] = m68k_read_memory_16(palette_address + i*2);
			/* for amiga speed: set delay by the number of cycles */
			if (config_amiga_speed)
				render_delay = (double)cycle_count / CPU_SPEED;
		}

		/* STEP 2: render image in memory via OpenGL */
		if (had_first_irq) {
			/* render game view without benson
			 * because benson is not updated before VBL IRQ, we don't want old image from double buffer
			 */
			emul_video(image, memory, palette, IMAGE_WIDTH, IMAGE_HEIGHT, IMAGE_DIWSTART, chipreg, 0, BENSON_AT_LINE, double_size);
		}
		rc = event_sdl();
		if (rc)
			break;
		render_opengl(image, config_video_filter);
		swap_sdl();

		/* STEP 3: execute interrupt at rate of 50Hz, render sound */
		current_time = SDL_GetTicks();
		diff = current_time - last_time;
		/* in case of timer glitch, execute IRQ only by maximum number of SOUND_CHUNKS */
		if (diff > 1000 * SOUND_CHUNKS / IRQ_RATE) {
			diff = 1000 * SOUND_CHUNKS / IRQ_RATE;
			last_time = current_time - 1000 * SOUND_CHUNKS / IRQ_RATE;
		}
		while (diff > 1000 / IRQ_RATE) {
			execute_cpu(3, NULL, NULL);
			execute_cpu(4, NULL, NULL);
			had_first_irq = 1;
			/* render benson without game view
			 * because we only got benson refreshed during VBL IRQ
			 */
			emul_video(image, memory, palette, IMAGE_WIDTH, IMAGE_HEIGHT, IMAGE_DIWSTART, chipreg, BENSON_AT_LINE, IMAGE_HEIGHT, double_size);
			/* render sound to sound buffer
			 * buffer pointer read and write is atomic, so no locking required!
			 */
			space = (sound_buffer_readp - sound_buffer_writep - 1 + sound_buffer_size) % sound_buffer_size;
			if (space < SOUND_SAMPLERATE / IRQ_RATE) {
#ifdef DEBUG_SOUND_BUFFERING
				fprintf(stderr, "sound buffer overflow\n");
#endif
				length = space;
			} else
				length = SOUND_SAMPLERATE / IRQ_RATE;
#ifdef DEBUG_SOUND_BUFFERING
			printf("can write %d, write %d\n", space, length);
			static int cnt = 0;
			int s;
			for (s = 0; s < length; s++) {
				sound_buffer[(sound_buffer_writep + s) % sound_buffer_size]
					= sin(2 * M_PI * 999 * cnt / SOUND_SAMPLERATE)
					* sin(2 * M_PI * 21 * cnt / SOUND_SAMPLERATE)
				 	* sin(2 * M_PI * 0.5 * cnt / SOUND_SAMPLERATE);
				 cnt++;
			}
#else
			render_sound(memory, sound_buffer, sound_buffer_size, sound_buffer_writep, length, config_audio_filter);
#endif
			sound_buffer_writep = (sound_buffer_writep + length) % sound_buffer_size;
			diff -= 1000 / IRQ_RATE;
			last_time += 1000 / IRQ_RATE;

			/* count down render delay */
			if (render_delay) {
				render_delay -= 1.0 / (double)IRQ_RATE;
				if (render_delay < 0.0)
					render_delay = 0.0;
			}
		}
	}
}

int main(int argc, char *argv[])
{
	int rc;
	int sdl_sound_chunk;

	home_dir = getenv("HOME");
	if (!home_dir)
		home_dir = "";

	rc = parse_args(argc, argv);
	if (rc)
		return 0;

	/* allocate image */
	image = calloc(IMAGE_WIDTH * IMAGE_HEIGHT * ((double_size) ? 4 : 1), 3);
	if (!image) {
		fprintf(stderr, "Failed to allocate image buffer\n");
		goto done;
	}

	if ((SOUND_SAMPLERATE % IRQ_RATE)) {
		fprintf(stderr, "Sample rate must be a multiple of IRQ rate, please fix!\n");
		goto done;
	}
	if ((1000 % IRQ_RATE)) {
		fprintf(stderr, "1000 (Ticks per second) rate must be a multiple of IRQ rate, please fix!\n");
		goto done;
	}

	/* calculate SDL chunk size for audio
	 * it must be a power of two, but not more than the chunk size for each IRQ!
	 */
	for (sdl_sound_chunk = 2; sdl_sound_chunk <= (SOUND_SAMPLERATE / IRQ_RATE); sdl_sound_chunk <<= 1)
		;
	sdl_sound_chunk >>= 1;
//	printf("samples per IRQ = %d, samples per SDL audio = %d\n", SOUND_SAMPLERATE / IRQ_RATE, sdl_sound_chunk); exit(0);

	/* allocate sound buffers */
	sound_buffer_size = SOUND_SAMPLERATE / IRQ_RATE * SOUND_CHUNKS;
	sound_buffer = calloc(sound_buffer_size, sizeof(*sound_buffer));
	if (!sound_buffer) {
		fprintf(stderr, "Failed to allocate image buffer\n");
		goto done;
	}

	/* allocate memory */
	memory = calloc(MEMORY_SIZE, 1);
	if (!memory) {
		fprintf(stderr, "Failed to allocate cpu's memory\n");
		goto done;
	}
	chipreg = calloc(IOSIZE, 1);
	if (!chipreg) {
		fprintf(stderr, "Failed to allocate chip register\n");
		goto done;
	}

	/* init cpu code */
	execute_init(MEMORY_SIZE, memory, chipreg);

	/* load binary */
	mercenary_load();

	/* patch some stuff */
	mercenary_patch();

	/* init SDL and OpenGL */
	rc = init_sdl(argv[0], SCREEN_WIDTH, SCREEN_HEIGHT, SOUND_SAMPLERATE, sdl_sound_chunk);
	if (rc < 0)
		goto done;
	rc = init_opengl((double_size) ? IMAGE_WIDTH * 2 : IMAGE_WIDTH, (double_size) ? IMAGE_HEIGHT * 2 : IMAGE_HEIGHT);
	if (rc < 0)
		goto done;
	resize_opengl(SCREEN_WIDTH, SCREEN_HEIGHT);

	/* init audio filter */
	sound_init_filter(SOUND_SAMPLERATE);

	/* start cpu */
	reset_cpu();

	printf("**********************************\n");
	printf("* Welcome to Mercenary Reloaded! *\n");
	printf("**********************************\n\n");
	printf("Press CTRL + cursor keys to select inventory or pickup/drop item.\n");
	printf("Press CTRL + f to toggle full screen.\n");
	printf("Press CTRL + s to toggle rendering speed.\n");
	printf("Press CTRL + v to toggle video filter.\n");
	printf("Press CTRL + a to toggle audio filter.\n");
	printf("Press CTRL + c to exit game.\n\n");
	printf("Use '--help' as command line option for configuration settings.\n\n");

	/* run game */
	main_loop();

done:
	exit_opengl();
	exit_sdl();

	if (chipreg)
		free(chipreg);
	if (memory)
		free(memory);
	if (sound_buffer)
		free(sound_buffer);
	if (image)
		free(image);

	return 0;
}

void audio_sdl(float *data, int length)
{
	int fill, s;

	/* read sound from sound buffer
	 * buffer pointer read and write is atomic, so no locking required!
	 */
	fill = (sound_buffer_writep - sound_buffer_readp + sound_buffer_size) % sound_buffer_size;
	if (fill < length) {
#ifdef DEBUG_SOUND_BUFFERING
		fprintf(stderr, "sound buffer underrun\n");
#endif
		/* correct read pointer as if the buffer would have 'length' of samples stored inside */
		sound_buffer_readp = (sound_buffer_readp + fill - length + sound_buffer_size) % sound_buffer_size;
	}
	for (s = 0; s < length; s++)
		*data++ = sound_buffer[(sound_buffer_readp + s) % sound_buffer_size];
#ifdef DEBUG_SOUND_BUFFERING
	printf("fill %d = %.4f\n", length, sound_buffer[sound_buffer_readp]);
#endif
	sound_buffer_readp =(sound_buffer_readp + length) % sound_buffer_size;
}

static int shift = 0, ctrl = 0;

void keyboard_sdl(int down, SDL_Keycode sym)
{
	switch (sym) {
	case SDLK_LCTRL:
	case SDLK_RCTRL:
		ctrl = down;
		break;
	case SDLK_v:
		if (down && ctrl) {
			config_video_filter ^= 1;
			printf("video filter: %s\n", (config_video_filter) ? "on" : "off");
		}
		break;
	case SDLK_a:
		if (down && ctrl) {
			config_audio_filter ^= 1;
			printf("audio filter: %s\n", (config_audio_filter) ? "on" : "off");
		}
		break;
	case SDLK_s:
		if (down && ctrl) {
			config_amiga_speed ^= 1;
			printf("amiga speed: %s\n", (config_amiga_speed) ? "original" : "full");
		}
		break;
	case SDLK_c:
		if (down)
			quit = 1;
		break;
	}

	switch (sym) {
	case SDLK_LSHIFT:
		set_key("LSH", down);
		shift = down;
		break;
	case SDLK_RSHIFT:
		set_key("RSH", down);
		shift = down;
		break;
	case SDLK_LEFT:
		if (shift && down) {
			set_key("LF", down);
			set_joystick(-1, -1, -1, -1, -1);
			break;
		}
		set_key("LF", 0);
		set_joystick(down, -1, -1, -1, -1);
		break;
	case SDLK_RIGHT:
		if (shift && down) {
			set_key("RT", down);
			set_joystick(-1, -1, -1, -1, -1);
			break;
		}
		set_key("RT", 0);
		set_joystick(-1, down, -1, -1, -1);
		break;
	case SDLK_UP:
		if (shift && down) {
			set_key("UP", down);
			set_joystick(-1, -1, -1, -1, -1);
			break;
		}
		set_key("UP", 0);
		set_joystick(-1, -1, down, -1, -1);
		break;
	case SDLK_DOWN:
		if (shift && down) {
			set_key("DN", down);
			set_joystick(-1, -1, -1, -1, -1);
			break;
		}
		set_key("DN", 0);
		set_joystick(-1, -1, -1, down, -1);
		break;
	case SDLK_END:
		set_joystick(-1, -1, -1, -1, down);
		break;

	case SDLK_a: set_key("A", down); break;
	case SDLK_b: set_key("B", down); break;
	case SDLK_c: set_key("C", down); break;
	case SDLK_d: set_key("D", down); break;
	case SDLK_e: set_key("E", down); break;
	case SDLK_f: set_key("F", down); break;
	case SDLK_g: set_key("G", down); break;
	case SDLK_h: set_key("H", down); break;
	case SDLK_i: set_key("I", down); break;
	case SDLK_j: set_key("J", down); break;
	case SDLK_k: set_key("K", down); break;
	case SDLK_l: set_key("L", down); break;
	case SDLK_m: set_key("M", down); break;
	case SDLK_n: set_key("N", down); break;
	case SDLK_o: set_key("O", down); break;
	case SDLK_p: set_key("P", down); break;
	case SDLK_q: set_key("Q", down); break;
	case SDLK_r: set_key("R", down); break;
	case SDLK_s: set_key("S", down); break;
	case SDLK_t: set_key("T", down); break;
	case SDLK_u: set_key("U", down); break;
	case SDLK_v: set_key("V", down); break;
	case SDLK_w: set_key("W", down); break;
	case SDLK_x: set_key("X", down); break;
	case SDLK_y: set_key("Y", down); break;
	case SDLK_z: set_key("Z", down); break;

	case SDLK_0: set_key("0", down); break;
	case SDLK_1: set_key("1", down); break;
	case SDLK_2: set_key("2", down); break;
	case SDLK_3: set_key("3", down); break;
	case SDLK_4: set_key("4", down); break;
	case SDLK_5: set_key("5", down); break;
	case SDLK_6: set_key("6", down); break;
	case SDLK_7: set_key("7", down); break;
	case SDLK_8: set_key("8", down); break;
	case SDLK_9: set_key("9", down); break;

	case SDLK_KP_0: set_key("NP0", down); break;
	case SDLK_KP_1: set_key("NP1", down); break;
	case SDLK_KP_2: set_key("NP2", down); break;
	case SDLK_KP_3: set_key("NP3", down); break;
	case SDLK_KP_4: set_key("NP4", down); break;
	case SDLK_KP_5: set_key("NP5", down); break;
	case SDLK_KP_6: set_key("NP6", down); break;
	case SDLK_KP_7: set_key("NP7", down); break;
	case SDLK_KP_8: set_key("NP8", down); break;
	case SDLK_KP_9: set_key("NP9", down); break;
	case SDLK_KP_DIVIDE: set_key("NPDIV", down); break;
	case SDLK_KP_MULTIPLY: set_key("NPMUL", down); break;
	case SDLK_KP_MINUS: set_key("NPSUB", down); break;
	case SDLK_KP_PLUS: set_key("NPADD", down); break;
	case SDLK_KP_PERIOD: set_key("NPDEL", down); break;
	// NPLPAREN and NPRPAREN are not emulated

	case SDLK_F1: set_key("F1", down); break;
	case SDLK_F2: set_key("F2", down); break;
	case SDLK_F3: set_key("F3", down); break;
	case SDLK_F4: set_key("F4", down); break;
	case SDLK_F5: set_key("F5", down); break;
	case SDLK_F6: set_key("F6", down); break;
	case SDLK_F7: set_key("F7", down); break;
	case SDLK_F8: set_key("F8", down); break;
	case SDLK_F9: set_key("F9", down); break;
	case SDLK_F10: set_key("F10", down); break;

	case SDLK_SPACE: set_key("SPC", down); break;
	case SDLK_BACKSPACE: set_key("BS", down); break;
	case SDLK_TAB: set_key("TAB", down); break;
	case SDLK_KP_ENTER: set_key("ENT", down); break;
	case SDLK_RETURN: set_key("RET", down); break;
	case SDLK_ESCAPE: set_key("ESC", down); break;
	case SDLK_DELETE: set_key("DEL", down); break;
	case SDLK_INSERT: set_key("HELP", down); break;

	}
}

uint16_t emulate_io_read(uint32_t address)
{
	uint16_t value = 0xffff;

	/* joystick and fire button */
	if (address == 0xbfe000 || address == 0xdff00c)
		value &= emulate_joystick_read(address);
	/* keyboard */
	if (address == 0xbfec00 || address == 0xbfed00 || address == 0xbfee00)
		value &= emulate_keyboard_read(address);
	/* diskette */
	if (address == 0xbfd100 || address == 0xbfe000 || address == 0xdff01a || address == 0xdff01e)
		value &= emulate_disk_read(address);

	return value;
}

void emulate_io_write(uint32_t address, uint16_t value)
{
	/* dmacon and sound registers */
	if (address == 0xdff096 || (address >= 0xdff0a0 && address <= 0xdff0df))
		emulate_sound_write(address, value, SOUND_SAMPLERATE);
	if (address == 0xbfd100 || (address >= 0xdff020 && address <= 0xdff024))
		emulate_disk_write(address, value);
}

/* two tracks with 0x1820 words of length */
static uint8_t game_save[2][0x1820 << 1];
static int last_track = 0;

/* game reads track with saved game */
void disk_read(int track, int __attribute__((unused)) side, uint32_t data, uint16_t length)
{
	if (length > sizeof(game_save[0])) {
		fprintf(stderr, "loading game state failed, because length exceeds game save data size, please fix!\n");
		return;
	}

	/* if even track is selected, load game */
	if (!(track & 1)) {
		char filename[256];
		int gamesave_num = (track - 2) >> 1;
		int got;
		FILE *fp;

		memset(game_save, 0, sizeof(game_save)); /* clear so make the game fail, if we fail */
		sprintf(filename, "%s/%s/%d.gamesave", home_dir, config_gamesave_dir, gamesave_num);
		fp = fopen(filename, "r");
		if (!fp) {
fail:
			fprintf(stderr, "failed to load game from '%s'\n", filename);
			goto copy;
		}
		got = fread(game_save, sizeof(game_save[0]), 2, fp);
		fclose(fp);
		if (got != 2)
			goto fail;
	}

copy:
	/* copy track */
	memcpy(memory + data, game_save[track & 1], length /* sizeof(game_save[0])*/);
}

/* game writes track with saved game */
void disk_write(int track, int __attribute__((unused)) side, uint32_t data, uint16_t length)
{
	/* skip sync info that is provided by game and only relevant for a real disk track */
	data += 0x200;
	length -= 0x200;

	if (length != sizeof(game_save[0])) {
		fprintf(stderr, "saving game state failed, because length of data does not match, please fix!\n");
		return;
	}

	/* don't save if last track is the same, because disk is written on both sides with the same data */
	if (track == last_track) {
		if (memcmp(memory + data, game_save[track & 1], length)) {
			fprintf(stderr, "saving game data on other side of the disk is different, please fix!\n");
		}
		return;
	}
	last_track = track;

	/* save game data */
	memcpy(game_save[track & 1], memory + data, length);

	/* if done with saving */
	if ((track & 1)) {
		char filename[256];
		int gamesave_num = (track - 2) >> 1;
		int wrote;
		FILE *fp;

		sprintf(filename, "%s/%s", home_dir, config_gamesave_dir);
		mkdir(filename, 0777);
		sprintf(filename + strlen(filename), "/%d.gamesave", gamesave_num);
		fp = fopen(filename, "w");
		if (!fp) {
fail:
			fprintf(stderr, "failed to save game to '%s'\n", filename);
			return;
		}
		printf("Game state saved to '%s'\n", filename);
		wrote = fwrite(game_save, sizeof(game_save[0]), 2, fp);
		fclose(fp);
		if (wrote != 2)
			goto fail;
	}
}
